/*
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <istream>
#include <concepts>
#include <memory>

namespace ext {

class istream {
	std::unique_ptr < std::istream > m_os;

public:
	istream ( std::basic_streambuf < char > * sb ) : m_os ( std::make_unique < std::istream > ( sb ) ) {
	}

	virtual ~istream ( ) noexcept = default;

protected:
	istream ( istream && rhs ) noexcept : m_os ( std::move ( rhs.m_os ) ) {
	}

	istream & operator = ( istream && rhs ) noexcept {
		m_os = std::move ( rhs.m_os );
		return * this;
	}

public:
	template < class T >
	requires requires { std::declval < std::istream > ( ) << std::declval < T > ( ); }
	friend istream & operator >> ( istream & os, const T & val ) {
		static_cast < std::istream & > ( os ) >> val;
		return os;
	}

	friend istream & operator >> ( istream & os, std::istream & ( * fn ) ( std::istream & ) ) {
		static_cast < std::istream & > ( os ) >> fn;
		return os;
	}

	friend istream & operator >> ( istream & os, std::ios_base & ( * fn ) ( std::ios_base & ) ) {
		static_cast < std::istream & > ( os ) >> fn;
		return os;
	}

protected:
	operator std::istream & ( ) & {
		return * m_os;
	}

	operator const std::istream & ( ) const & {
		return * m_os;
	}

public:

	int peek ( ) {
		return m_os->peek ( );
	}

	int get ( ) {
		return m_os->get ( );
	}

	istream & get ( char & ch ) {
		m_os->get ( ch );
		return * this;
	}

	istream & unget ( ) {
		m_os->unget ( );
		return * this;
	}

	istream & putback ( char ch ) {
		m_os->putback ( ch );
		return * this;
	}

	istream & read ( char * s, std::streamsize count ) {
		m_os->read ( s, count );
		return * this;
	}

	std::streampos tellg ( ) {
		return m_os->tellg ( );
	}

	istream & seekg ( std::streampos pos ) {
		m_os->seekg ( pos );
		return * this;
	}

	istream & seekg ( std::streamoff off, std::ios_base::seekdir dir ) {
		m_os->seekg ( off, dir );
		return * this;
	}

	istream & sync ( ) {
		m_os->sync ( );
		return * this;
	}

	std::streambuf * rdbuf ( ) const {
		return m_os->rdbuf ( );
	}

	std::streambuf * rdbuf ( std::streambuf * buf ) {
		return m_os->rdbuf ( buf );
	}

	void setstate ( std::ios_base::iostate state ) {
		m_os->setstate ( state );
	}

	void clear ( std::ios_base::iostate state = std::ios_base::goodbit ) {
		m_os->clear ( state );
	}

	bool good ( ) const {
		return m_os->good ( );
	}

	bool fail ( ) const {
		return m_os->fail ( );
	}

	bool eof ( ) const {
		return m_os->eof ( );
	}

	operator bool ( ) const {
		return m_os->operator bool ( );
	}

};

/**
 * \brief
 * Read into a constant string. The implementation tests whether exact same sequence represented as \p str is in the \p in stream. In case it is, it reads it. In case it is not the stream is not advanced and fail bit is set.
 *
 * \param in the input stream
 * \param str the string to test
 *
 * \return the in stream either with fail bit set if str is not in the stream or advanced by reading str.
 */
ext::istream & operator >> ( ext::istream & in, const std::string & str );

} /* namespace ext */

